#include <vector>
#include <cppcore/conversion/string.h>
#include <cppmicrohttpd/config.h>
#include <cppmicrohttpd/daemon.h>
#include <cppmicrohttpd/request/request.inl>

using namespace ::cppmicrohttpd;

daemon::daemon(
    uint16_t                port,
    const daemon_flags&     flags,
    const callback_flags&   callback)
    : _handle(nullptr, &MHD_stop_daemon)
{
    std::vector<MHD_OptionItem> options;
    // TODO add more
    options.emplace_back(MHD_OptionItem { MHD_OPTION_END, 0, nullptr });

    _handle.reset(MHD_start_daemon(
        flags.value,
        port,
        nullptr,
        nullptr,
        &daemon::mhd_access_handler_callback,
        this,
        MHD_OPTION_ARRAY, options.data(),
        MHD_OPTION_END));

    if (!_handle)
        throw exception("Unable to create daemon");
}

request_ptr_u daemon::create_request(
    MHD_Connection * const p_connection,
    const char * const p_url,
    const char * const p_method,
    const char * const p_version)
    { return std::make_unique<request>(p_connection, p_url, p_method, p_version); }

response_ptr_u daemon::create_response(
    request& p_request)
    { return create_default_response(p_request, true); }

response_ptr_u daemon::create_default_response(
    const request& p_request,
    bool p_error_only)
{
    using namespace ::std;

    if (p_request.error)
    {
        auto* http_ex = dynamic_cast<const http_exception*>(p_request.error.get());
        auto* core_ex = dynamic_cast<const ::cppcore::exception*>(p_request.error.get());
        if (http_ex)
        {
            return response::build_simple_html_response(
                p_request,
                http_ex->status,
                "Status "s + ::cppcore::to_string(http_ex->status),
                "Status "s + ::cppcore::to_string(http_ex->status),
                http_ex->print(::cppcore::exception::print_flags::empty()));
        }
        else if (core_ex)
        {
            return response::build_simple_html_response(
                p_request,
                500,
                "Status 500"s,
                "Status 500"s,
                core_ex->print(::cppcore::exception::print_flags::empty()));
        }
        else
        {
            return response::build_simple_html_response(
                p_request,
                500,
                "Status 500"s,
                "Status 500"s,
                "Unknown error. Check server error log.");
        }
    }

    return response::build_simple_html_response(
        p_request,
        404,
        "Status 404"s,
        "Status 404"s,
        "The requested URL could not be found on this server: "s + p_request.url);
}

int daemon::mhd_access_handler_callback(
    void * cls,
    struct MHD_Connection * connection,
    const char * url,
    const char * method,
    const char * version,
    const char * post_data,
    size_t * data_size,
    void ** con_cls)
{
    if (!cls)
        return MHD_NO;

    if (!con_cls)
        return MHD_NO;

    auto& daemon = *static_cast<::cppmicrohttpd::daemon*>(cls);



    /* create request object */
    try
    {
        if (!*con_cls)
        {
            auto tmp = daemon.create_request(connection, url, method, version);
            if (!tmp)
                throw exception("No request object was created");
            *con_cls = tmp.release();
            return MHD_YES;
        }
    }
    catch(const std::exception& ex)
    {
        cppmicrohttpd_log(error) << "Error while creating request: " << ex.what();
        return MHD_NO;
    }
    catch(...)
    {
        cppmicrohttpd_log(error) << "Error while creating request: unknown";
        return MHD_NO;
    }



    /* handle request data */
    auto& req = *static_cast<request*>(*con_cls);
    try
    {
        if (req.handle_post_pata(post_data, *data_size))
            return MHD_YES;
    }
    catch(const http_exception& ex)
    {
        cppmicrohttpd_log(warn) << "Error while handling post data: " << ex;
        req.set_error(ex);
        *data_size = 0; // Assume that all data has been processed
        return MHD_YES;
    }
    catch(const exception& ex)
    {
        cppmicrohttpd_log(warn) << "Error while handling post data: " << ex;
        req.set_error(ex);
        *data_size = 0; // Assume that all data has been processed
        return MHD_YES;
    }
    catch(std::exception& ex)
    {
        cppmicrohttpd_log(warn) << "Error while handling post data: " << ex.what();
        req.set_error(exception(ex.what()));
        *data_size = 0; // Assume that all data has been processed
        return MHD_YES;
    }
    catch(...)
    {
        cppmicrohttpd_log(warn) << "Error while handling post data: unknown";
        req.set_error(exception("Unknown error"));
        *data_size = 0; // Assume that all data has been processed
        return MHD_YES;
    }



    /* create response object */
    try
    {
        if (!req.response)
        {
            req.response = daemon.create_response(req);
            if (!req.response)
                throw exception("No response object was created");
            auto& res = *req.response;
            if (MHD_queue_response(req.connection, res.status, res.handle.get()) != MHD_YES)
                throw exception("Unable to enqueue response");
        }
        return MHD_YES;
    }
    catch(const std::exception& ex)
    {
        cppmicrohttpd_log(error) << "Error while creating response: " << ex.what();
        return MHD_NO;
    }
    catch(...)
    {
        cppmicrohttpd_log(error) << "Error while creating response: unknown";
        return MHD_NO;
    }
}
